; ==============================================================================
; DLL de morpion 3D  4x4x4x4
; Version 0.1
;
; avec FASM, par Sad Koudache 2007
; ==============================================================================

format PE GUI 4.0 DLL
entry DllEntryPoint
; stack 100h

include 'win32a.inc'

align 4

; ==============================================================================
section '.data' data readable writeable

; terrain de jeu, l o seront sauvegards les coups jous.
jeu	  db 199 + 48 + 1 dup(?)
jeu_size  equ 248

; nombre d'alignements possibles de 4 mmes symboles
resultat  dw  0
ali	  db  200 dup(?)
ali_size  equ 200

; contenu du terrain de jeu
VIDE	  equ 0
JOUEUR	  equ 1
MACHINE   equ 4
BORD	  equ 5

; joueur qui doit jouer au prochain tour
; prochain  db ?

; dfinition des diffrents joueurs

; les diffrents gains, notez qu'on encourage le gain de la machine, ainsi elle
; prfrera gagner que bloquer un gain de l'adversaire.
GAIN_J	  equ	2000
GAIN_M	  equ  10000

; note des diffrents alignements
notes	  dw  0,   30,	 61,  GAIN_J, \
		   40,	  0,	   0,	    0,	 \
		  120,	  0,	   0,	    0,	 \
	       GAIN_M,	  0,	   0,	    0

gain	  dw  0, GAIN_J, 0, 0, GAIN_M

; les directions  explorer
dir	  db  1, 6, 5, 4, 36, 34, 40, 44, 45, 46, 41, 48, 39, 35, 0

; pour afficher la ligne finale, on a besoin de la case dpart et de fin
case_dep_gain dd ?
case_arr_gain dd ?

; ==============================================================================
section '.code' code readable executable
align 4

proc DllEntryPoint hinstDLL, fdwReason, lpvReserved
   mov eax, TRUE
   ret
endp


; ==============================================================================
; vider_jeu
;
; permet de remettre tout le terrain de jeu  0 pour une nouvelle partie
; paramtres : aucun
; ==============================================================================

proc vider_jeu
    ; on met d'abord tout le terrain  BORD
    cld
    mov ecx, 248
    mov edi, jeu
    mov al, BORD
    rep stosb

    ; ensuite on met les cases jouables  VIDE
    xor   eax, eax
    mov   ecx, 4
  .bou_cx:
    xor   edx, edx
  .bou_dx:
    mov   byte [jeu + eax + edx + 61],	VIDE
    mov   byte [jeu + eax + edx + 101], VIDE
    mov   byte [jeu + eax + edx + 141], VIDE
    mov   byte [jeu + eax + edx + 181], VIDE

    inc   edx
    cmp   edx, 3
    jbe  .bou_dx

    add   eax, 5
    loop .bou_cx

   ret
endp   


; ==============================================================================
; nouvelle_partie
;
; permet de tout rinitialiser pour refaire une nouvelle partie
; paramtres : aucun
; type       : exporte
; ==============================================================================

proc nouvelle_partie
   call vider_jeu
   call initialisation_alignements

   mov dword [case_dep_gain], 0
   mov dword [case_arr_gain], 0

   ret
endp


proc get_resultat
   mov eax, dword [resultat]
   ret
endp


; ==============================================================================
; que_contient_la_case
;
; permet de rcuprer le contenu d'une case du jeu
; paramtres : case_demandee (de 61  200)
; type       : exporte
; ==============================================================================

proc que_contient_la_case   case_demandee:DWORD
   mov ebx, [case_demandee]
   xor eax, eax
   mov al, byte [jeu + ebx]

   ret
endp

; ==============================================================================
; cherche_alignements
;
; permet de prcalculer tous les alignements possibles de 4 pices par case
; paramtres : coup (= case_demandee de 61  200)
; type       : interne
; ==============================================================================

proc cherche_alignements coup:DWORD
   pushad
   xor eax, eax

  ; pour chaque direction, vrifier compter les possibilits sur
   mov esi, dir
 .debut:	      ; pour chaque direction
   mov al, byte [esi]

   xor edx, edx       ; case du jeu que l'on tudie
   mov edi, jeu
   add edi, DWORD [coup]

 .bou_edx:
   add edi, eax       ; incremente la direction
   cmp byte [edi], BORD
   je  .dir_suivante  ; fin d'alignement donc ligne incomplte, on teste la 
		      ; direction suivante

   inc edx
   cmp edx, 3
   jb .bou_edx

   ; on est arriv ici, c'est qu'un alignement de 4 pices est possible.
   mov ecx, 4
   mov edi, ali
   add edi, dword [coup]
 .maj_align:
   inc byte [edi]     ; on ajoute 1  chaque case traverse par l'alignement, 
   add edi, eax       ; la machine choisira ainsi les coups les
   loop .maj_align    ; plus prometteurs en terme d'alignements possibles.

 .dir_suivante:
   inc esi
   cmp byte [esi], 0
   jne .debut		; direction suivante TQ dir != 0 .

 .fin:
   popad
   ret
endp

; ==============================================================================
; initialisation_alignements
;
; permet de prcalculer tous les alignements possibles de 4 pices pour toutes
; les cases
; paramtres : coup (= case_demandee de 61  200)
; type       : interne
; ==============================================================================

proc initialisation_alignements
    pushad
    cld
    mov ecx, 200
    mov edi, ali
    xor al, al
    rep stosb

    mov edx, 61

   .boucle:
    cmp byte [jeu + edx], BORD
    jz .c_est_un_bord

    stdcall cherche_alignements, edx

   .c_est_un_bord:
    inc edx
    cmp edx, 200
    jb	.boucle

    popad
    ret
endp

; ==============================================================================
; evaluation
;
; permet de calculer la note stratgique d'une case (note d'alignement + note de
; position
; paramtres : coup (= case_demandee de 61  200)
;              tour (= constante MACHINE 4 ou JOUEUR 1)
; type       : interne
; ==============================================================================

proc evaluation  coup:DWORD, tour:DWORD
    local note:DWORD	 ; note cumule des diffrents alignements sur cette case.
    local gain_tour:DWORD

    pushad

    ; vrifier que l'on examine pas une case de bord (utile seulement pour mes tests) 
    ;  mettre en commentaires ensuite
    mov ebx, dword [coup]
    cmp byte [jeu + ebx], BORD
    jne .OK

    mov dword [note], -1
    jmp .ret
    ;--------------------------------------------------------------------------------

  .OK:
    mov dword [note], 0

    xor eax, eax
    mov edi, ali
    add edi, dword [coup]
    mov al, byte [edi]
    mov dword [note], eax    ; note contient au moins le nb d'align. possibles

    mov eax, dword [tour]    ; note d'1 aligt gagnant : 3 fois le pion du joueur
    add eax, dword [tour]    ; (car pion  jouer, n'est pas encore pos)
    add eax, dword [tour]
    mov dword [gain_tour], eax	; gain_tour contient 12 ou 3 selon le tour

    mov dword [case_dep_gain], 0 ; marquer les cases de gain  vide
    mov dword [case_arr_gain], 0

   ; pour chaque direction, vrifier compter les possibilits sur
    mov esi, dir
 .debut:		     ; pour chaque direction
    xor eax, eax
    mov al, byte [esi]
    mov edi, jeu
    add edi, dword [coup]

    ; remonter jusqu' la premire case dans la direction oppose.
    push edi			    ; sav pour calculs descendant et remontant

    xor ebx, ebx		    ; compteur
    xor edx, edx		    ; case du jeu que l'on tudie

 .compter_en_remontant:
    sub edi, eax
    cmp byte [edi], BORD
    je .suite_comptage

    add bl, byte [edi]
    jmp short .compter_en_remontant

 .suite_comptage:
    pop edi

 .compter_en_descendant:
    add edi, eax
    cmp byte [edi], BORD
    je .fin_comptage

    add bl, byte [edi]
    jmp short .compter_en_descendant

 .fin_comptage:
    cmp dword  [gain_tour], ebx
    je	.gain_trouve

    xor eax, eax			; rcuprer la valeur de l'alignt du tableau notes
    mov ax, word [notes + ebx * 2]	; et l'ajouter  la note cumule de cette case
    add dword [note], eax

 .dir_suivante:
   inc esi
   cmp byte [esi], 0
   jne .debut				; direction suivante TQ dir != 0 .

   mov dword [case_dep_gain], 0 	; marquer les cases de gain  vide
   mov dword [case_arr_gain], 0

   jmp short .ret			; On renvoie la somme des notes par alignements

  ; OK on a trouv le gain !
  .gain_trouve:
    ; rcuprer les cases dbut et de fin de l'alignement, pour tracer la ligne dans l'interface utilisateur
    sub edi, jeu		   ; transformer la coordonne physique du jeu  en indice de case
    sub edi, eax		   ; on remonte dans la direction courante pour rcuprer les cases
    mov dword [case_arr_gain], edi ; marquer les cases de gain  vide
    sub edi, eax
    sub edi, eax
    sub edi, eax
    mov dword [case_dep_gain], edi

    mov esi, notes		       ; rcuprer val dans le tableau des notes
    mov ax, word [esi + ebx * 2] ; pour acclerer, on pouvait retourner GAIN_M=10000

    mov dword [note], eax	       ; Attn car EAX perdu aprs le popad 

 .ret:
    popad
    mov eax, dword [note] ; eax, contient le rsultat de l'valuation de la case
    ret
endp


; ==============================================================================
; joueur_joue
;
; permet au joueur de placer un pion
; paramtres : coup (= case_demandee de 61  200)
;              cote (= constante MACHINE 4 ou JOUEUR 1)
; type       : exporte
; ==============================================================================

proc joueur_joue coup:DWORD, cote:BYTE
    local fin_partie:DWORD

    push ebx eax edx

    mov dword [fin_partie], 0		; = partie en cours

    call calculer_cases_libres		; Si nb cases libres <= 1 alors probablement nulle.
    cmp  eax, 1
    jg .suite

    mov dword [fin_partie], 2		; = nulle, sauf si gain trouv, dans les lignes ci dessous

   .suite:
    mov dl, byte [cote] 		; on vrifie que la case est bien vide
    mov ebx, dword [coup]
    cmp byte [jeu + ebx], VIDE
    jnz .ret

    stdcall evaluation, ebx, JOUEUR	; garde et restore tous les registres, rsultat dans EAX

    mov byte [jeu + ebx], dl		; placer le pion du joueur

    cmp eax, GAIN_J			; la note renvoie-t-elle un gain du joueur
    jnz .ret				; non : on quitte

    mov dword [fin_partie], 1		; oui, on flague le gain du joueur

  .ret:
    pop edx eax ebx
    mov eax, dword [fin_partie]

    ret
endp


; ==============================================================================
; calculer_cases_libres
;
; permet de compter le nb de cases vides sur le jeu pour la dtection des nulles
;
; paramtres :
; rsultat   : EAX = nb cases vides
; type       : exporte
; ==============================================================================

proc calculer_cases_libres		; EAX contient le nombre de cases vides
    push ecx ebx

    mov ebx, 60
    mov ecx, 140
    xor eax, eax
  .b:
    inc ebx
    cmp byte [jeu + ebx], VIDE
    jne .suite_b

    inc eax				; +1 case de libre

  .suite_b:
    loop .b

  .ret:
    pop ebx ecx
    ret
endp

; ==============================================================================
; machine_joue
;
; Rflexion de la machine, elle value toutes les cases du jeu et choisit le
;                          meilleur coup.
; paramtres : coup_resultat = int *, retourne le coup choisi par la machine)
;              eax  = rsultat de la partie (0=en cours, 1=gain machine, 2 = gain joueur)
;              cote = constante MACHINE 4 ou JOUEUR 1)
; type       : exporte
; ==============================================================================

proc  machine_joue coup_resultat:DWORD
    local fin_partie:DWORD

    pushad

    mov  dword [fin_partie], 2	       ; match nul, au cas o toutes les cases seraient remplies
    mov  edx, -30000		       ; meilleur score, initialis  0, EAX contiendra la valeur de la case
    mov  ebx, 61		       ; indice de boucle des cases  valuer
    mov  edi, [coup_resultat]	       ; attention [coup_resultat] contient un int *
    mov  dword [edi], 0

  .boucle_case:
    cmp byte [jeu + ebx], VIDE
    jnz .prochaine_case

    stdcall evaluation, ebx, MACHINE
    cmp edx, eax
    jge .prochaine_case

    ;on a trouv un meilleur coup, donc on le garde, ainsi que la note
    mov DWORD [edi], ebx
    mov edx,  eax

    ;on vrifie si ce n'est pas un gain (car pas la peine de continuer)
    cmp edx, GAIN_M
    je	.sortie_gain

  .prochaine_case:
    inc ebx
    cmp ebx, 200
    jb	.boucle_case

  .sortie_boucle:
    ; vrifier si au moins un passage dans la boucle a t ralis, sinon on quitte, type de fin partie = -1
    cmp edx, -30000
    je .ret

    mov  dword [fin_partie], 0		   ; partie en cours, TQ non prouv le contraire
    jmp short .test_nulle

  .sortie_gain: 			   ; gain trouv, pas la peine de tester si match nul !
    mov  dword [fin_partie], 1
    jmp short .suite

  .test_nulle:
    call calculer_cases_libres		   ; Si nb cases libres <= 1 et pas de gain alors c'est match nul.
    cmp  eax, 1
    jg .suite

    mov dword [fin_partie], 2		   ; = nulle, sauf si gain trouv, dans les lignes ci dessous

  .suite:
    mov ebx, DWORD [edi]		   ; rcupration du meilleur coup
    mov byte [jeu + ebx], MACHINE	   ; et mise  jour du jeu

  .ret:
    popad
    mov eax, dword [fin_partie]
    ret
endp


; ==============================================================================
; donner_cases_de_gain
;
; donne au programme appelant la case de dpart et la case de fin, le programme
;  appelant calculera les points traverss et affichera une droite.
;
; paramtres : pCaseDepart = int *, case dpart
;              pCaseFin    = int *, case fin
; type       : exporte
; ==============================================================================

proc donner_cases_de_gain pCaseDepart:DWORD, pCaseFin:DWORD
    push esi eax

    ; case dpart
    mov esi, [pCaseDepart]
    mov eax, dword [case_dep_gain]
    mov dword [esi], eax

    ; case de fin
    mov esi, [pCaseFin]
    mov eax, dword [case_arr_gain]
    mov dword [esi], eax

    pop eax esi
    ret
endp

; ==============================================================================

data import
 ; section '.idata' import data readable writeable
  library kernel32,'KERNEL32.DLL'

  import kernel32,\
    GlobalAlloc,       'GlobalAlloc',\
    GlobalFree,        'GlobalFree'
    ; OutputDebugString, 'OutputDebugStringA'
end data


section '.edata' export data readable
 export 'M3D.DLL',\
	 nouvelle_partie,      'nouvelle_partie', \   
	 get_resultat,	       'get_resultat',\
	 que_contient_la_case, 'que_contient_la_case',\
	 evaluation,	       'evaluation',\
	 joueur_joue,	       'joueur_joue',\	     
	 machine_joue,	       'machine_joue',\      
	 donner_cases_de_gain, 'donner_cases_de_gain'

; --------------------------------------------------------------------------------
; section .reloc obligatoire, sinon appel  fonctions plante, ex : nouvelle_partie
section '.reloc' fixups data discardable

section '.rsrc' resource data readable
  directory   RT_VERSION,versions

  resource versions,\
	   1, LANG_NEUTRAL,version

  versioninfo version,VOS__WINDOWS32,VFT_APP,VFT2_UNKNOWN,LANG_FRENCH+SUBLANG_DEFAULT,0,\
	      'FileDescription',  'M3D - Moteur en ASM 32 bits avec FASM',\
	      'LegalCopyright',   'Sad Koudache, 2007',\
	      'FileVersion',	  '0.1',\
	      'ProductVersion',   '0.1',\
	      'OriginalFilename', 'M3D.DLL'


